void    *memccpy(void *, void *, int, long);
void    *memchr(void *, int, long);
int      memcmp(void *, void *, long);
void    *memcpy(void *, void *, long);
void    *memmove(void *, void *, long);
void    *memset(void *, int, long);
char    *strcat(char *, char *);
char    *strchr(char *, int);
int      strcmp(char *, char *);
int      strcoll(char *, char *);
char    *strcpy(char *, char *);
long    strcspn(char *, char *);
char    *strdup(char *);
char    *strerror(int);
long    strlen(char *);
char    *strncat(char *, char *, long);
int      strncmp(char *, char *, long);
char    *strncpy(char *, char *, long);
char    *strpbrk(char *, char *);
char    *strrchr(char *, int);
long    strspn(char *, char *);
char    *strstr(char *, char *);
char    *strtok(char *, char *);
char    *strtok_r(char *, char *, char **);
long    strxfrm(char *, char *, long);
