long      a64l(char *);
void      abort(void);
int       abs(int);
int       atexit(void (*)(void));
int       atoi(char *);
long      atol(char *);
void     *bsearch(void *, void *, long, long, int (*)(void *, void *));
void     *calloc(long, long);
void      exit(int);
void      free(void *);
char     *getenv(char *);
int       getsubopt(char **, char **, char **);
int       grantpt(int);
char     *initstate(unsigned int, char *, long);
long      jrand48(unsigned short[3]);
char     *l64a(long);
long      labs(long);
void      lcong48(unsigned short[7]);
long      lrand48(void);
void     *malloc(long);
int       mblen(char *, long);
char     *mktemp(char *);
int       mkstemp(char *);
long      mrand48(void);
long      nrand48(unsigned short [3]);
char     *ptsname(int);
int       putenv(char *);
void      qsort(void *, long, long, int (*)(void *, void *));
int       rand(void);
int       rand_r(unsigned int *);
long      random(void);
void     *realloc(void *, long);
char     *realpath(char *, char *);
unsigned short
          seed48(unsigned short[3]);
void      setkey(char *);
char     *setstate(char *);
void      srand(unsigned int);
void      srand48(long);
void      srandom(unsigned);
long      strtol(char *, char **, int);
unsigned long
          strtoul(char *, char **, int);
int       system(char *);
int       unlockpt(int);
